@@ -2,10 +2,22 @@
 
 """Resolve classes."""
 
+import inspect
+from operator import attrgetter
 from textwrap import dedent
-from typing import Any, Collection, Generic, Iterable, Mapping, Optional, Set, Type, TypeVar, Union
+from typing import (
+    Any, Callable, Collection, Dict, Generic, Iterable, Iterator, Mapping, Optional, Set, TYPE_CHECKING, Type,
+    TypeVar, Union,
+)
+
+if TYPE_CHECKING:
+    import click
 
 __all__ = [
+    'InstOrType',
+    'Lookup',
+    'LookupType',
+    'LookupOrType',
     'Hint',
     'HintType',
     'HintOrType',
@@ -16,9 +28,17 @@ __all__ = [
 ]
 
 X = TypeVar('X')
-Hint = Union[None, str, X]
+Y = TypeVar('Y')
+
+InstOrType = Union[X, Type[X]]
+
+Lookup = Union[str, X]
+LookupType = Lookup[Type[X]]
+LookupOrType = Lookup[InstOrType[X]]
+
+Hint = Optional[Lookup[X]]
 HintType = Hint[Type[X]]
-HintOrType = Hint[Union[X, Type[X]]]
+HintOrType = Hint[InstOrType[X]]
 
 
 class Resolver(Generic[X]):
@@ -29,7 +49,9 @@ class Resolver(Generic[X]):
     #: The shared suffix fo all classes derived from the base class
     suffix: str
     #: The mapping from normalized class names to the classes indexed by this resolver
-    lookup_dict: Mapping[str, Type[X]]
+    lookup_dict: Dict[str, Type[X]]
+    #: The mapping from synonyms to the classes indexed by this resolver
+    synonyms: Dict[str, Type[X]]
 
     def __init__(
         self,
@@ -54,11 +76,42 @@ class Resolver(Generic[X]):
         if suffix is None:
             suffix = normalize_string(base.__name__)
         self.suffix = suffix
-        self.synonyms = synonyms
-        self.lookup_dict = {
-            self.normalize_cls(cls): cls
-            for cls in classes
-        }
+        self.synonyms = dict(synonyms or {})
+        self.lookup_dict = {}
+        for cls in classes:
+            self.register(cls)
+
+    def register(self, cls: Type[X], synonyms: Optional[Iterable[str]] = None, raise_on_conflict: bool = True) -> None:
+        """Register an additional class with this resolver.
+
+        :param cls: The class to register
+        :param synonyms: An optional iterable of synonyms to add for the class
+        :param raise_on_conflict: Determines the behavior when a conflict is encountered on either
+            the normalized class name or a synonym. If true, will raise an exception. If false, will
+            simply disregard the entry.
+
+        :raises KeyError: If ``raise_on_conflict`` is true and there's a conflict in either the class
+            name or a synonym name.
+        """
+        key = self.normalize_cls(cls)
+        if key in self.lookup_dict:
+            if raise_on_conflict:
+                raise KeyError(f'This resolver already contains a class with key {key}: {self.lookup_dict[key]}')
+            else:
+                return
+
+        self.lookup_dict[key] = cls
+        for synonym in synonyms or []:
+            if synonym in self.synonyms:
+                if raise_on_conflict:
+                    raise KeyError(f'This resolver already contains synonym {synonym} for {self.synonyms[synonym]}')
+                else:
+                    continue
+            self.synonyms[synonym] = cls
+
+    def __iter__(self) -> Iterator[Type[X]]:
+        """Return an iterator over the indexed classes sorted by name."""
+        return iter(sorted(self.lookup_dict.values(), key=attrgetter('__name__')))
 
     @classmethod
     def from_subclasses(cls, base: Type[X], *, skip: Optional[Collection[Type[X]]] = None, **kwargs) -> 'Resolver':
@@ -103,7 +156,16 @@ class Resolver(Generic[X]):
             suffix=self.suffix,
         )
 
-    def make(self, query: Hint[Union[X, Type[X]]], pos_kwargs: Optional[Mapping[str, Any]] = None, **kwargs) -> X:
+    def signature(self, query: Hint[Type[X]]) -> inspect.Signature:
+        """Get the signature for the given class via :func:`inspect.signature`."""
+        cls = self.lookup(query)
+        return inspect.signature(cls)
+
+    def supports_argument(self, query: Hint[Type[X]], parameter_name: str) -> bool:
+        """Determine if the class constructor supports the given argument."""
+        return parameter_name in self.signature(query).parameters
+
+    def make(self, query: HintOrType[X], pos_kwargs: Optional[Mapping[str, Any]] = None, **kwargs) -> X:
         """Instantiate a class with optional kwargs."""
         if query is None or isinstance(query, (str, type)):
             cls: Type[X] = self.lookup(query)
@@ -112,21 +174,46 @@ class Resolver(Generic[X]):
         # An instance was passed, and it will go through without modification.
         return query
 
-    def get_option(self, *flags: str, default: Optional[str] = None, **kwargs):
+    def make_from_kwargs(
+        self,
+        data: Mapping[str, Any],
+        key: str,
+        *,
+        kwargs_suffix: str = "kwargs",
+        **o_kwargs,
+    ) -> X:
+        """Instantiate a class, by looking up query/pos_kwargs from a dictionary.
+
+        :param data: A dictionary that contains entry ``key`` and entry ``{key}_{kwargs_suffix}``.
+        :param key: The key in the dictionary whose value will be put in the ``query`` argument of :func:`make`.
+        :param kwargs_suffix: The suffix after ``key`` to look up the data. For example, if ``key='model'``
+            and ``kwargs_suffix='kwargs'`` (the default value), then the kwargs from :func:`make` are looked up
+            via ``data['model_kwargs']``.
+        :param o_kwargs: Additional kwargs to be passed to :func:`make`
+        :returns: An instance of the X datatype parametrized by this resolver
+        """
+        query = data.get(key, None)
+        pos_kwargs = data.get(f"{key}_{kwargs_suffix}", {})
+        return self.make(query=query, pos_kwargs=pos_kwargs, **o_kwargs)
+
+    def get_option(self, *flags: str, default: Hint[Type[X]] = None, as_string: bool = False, **kwargs):
         """Get a click option for this resolver."""
         if default is None:
             if self.default is None:
-                raise ValueError
-            default = self.normalize_cls(self.default)
+                raise ValueError('no default given either from resolver or explicitly')
+            default = self.default
+        else:
+            default = self.lookup(default)
+        default = self.normalize_cls(default)
 
         import click
 
         return click.option(
             *flags,
-            type=click.Choice(list(self.lookup_dict)),
+            type=click.Choice(list(self.lookup_dict), case_sensitive=False),
             default=default,
             show_default=True,
-            callback=_make_callback(self.lookup),
+            callback=None if as_string else _make_callback(self.lookup),
             **kwargs,
         )
 
@@ -255,8 +342,8 @@ def normalize_string(s: str, *, suffix: Optional[str] = None) -> str:
     return s
 
 
-def _make_callback(f):
-    def _callback(_, __, value):
+def _make_callback(f: Callable[[X], Y]) -> Callable[['click.Context', 'click.Parameter', X], Y]:
+    def _callback(_ctx: 'click.Context', _param: 'click.Parameter', value: X) -> Y:
         return f(value)
 
     return _callback
